@@ -1,28 +1,67 @@
 package com.genymobile.scrcpy.control;
 
+import com.genymobile.scrcpy.AndroidVersions;
 import com.genymobile.scrcpy.AsyncProcessor;
 import com.genymobile.scrcpy.CleanUp;
+import com.genymobile.scrcpy.Options;
 import com.genymobile.scrcpy.device.Device;
-import com.genymobile.scrcpy.util.Ln;
+import com.genymobile.scrcpy.device.DeviceApp;
+import com.genymobile.scrcpy.device.DisplayInfo;
 import com.genymobile.scrcpy.device.Point;
 import com.genymobile.scrcpy.device.Position;
+import com.genymobile.scrcpy.device.Size;
+import com.genymobile.scrcpy.util.Ln;
+import com.genymobile.scrcpy.util.LogUtils;
+import com.genymobile.scrcpy.video.SurfaceCapture;
+import com.genymobile.scrcpy.video.VirtualDisplayListener;
+import com.genymobile.scrcpy.wrappers.ClipboardManager;
 import com.genymobile.scrcpy.wrappers.InputManager;
 import com.genymobile.scrcpy.wrappers.ServiceManager;
 
 import android.content.Intent;
 import android.os.Build;
 import android.os.SystemClock;
+import android.util.Pair;
 import android.view.InputDevice;
 import android.view.KeyCharacterMap;
 import android.view.KeyEvent;
 import android.view.MotionEvent;
 
 import java.io.IOException;
+import java.util.List;
+import java.util.concurrent.ExecutorService;
 import java.util.concurrent.Executors;
 import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicReference;
 
-public class Controller implements AsyncProcessor {
+public class Controller implements AsyncProcessor, VirtualDisplayListener {
+
+    /*
+     * For event injection, there are two display ids:
+     *  - the displayId passed to the constructor (which comes from --display-id passed by the client, 0 for the main display);
+     *  - the virtualDisplayId used for mirroring, notified by the capture instance via the VirtualDisplayListener interface.
+     *
+     * (In case the ScreenCapture uses the "SurfaceControl API", then both ids are equals, but this is an implementation detail.)
+     *
+     * In order to make events work correctly in all cases:
+     *  - virtualDisplayId must be used for events relative to the display (mouse and touch events with coordinates);
+     *  - displayId must be used for other events (like key events).
+     *
+     * If a new separate virtual display is created (using --new-display), then displayId == Device.DISPLAY_ID_NONE. In that case, all events are
+     * sent to the virtual display id.
+     */
+
+    private static final class DisplayData {
+        private final int virtualDisplayId;
+        private final PositionMapper positionMapper;
+
+        private DisplayData(int virtualDisplayId, PositionMapper positionMapper) {
+            this.virtualDisplayId = virtualDisplayId;
+            this.positionMapper = positionMapper;
+        }
+    }
 
     private static final int DEFAULT_DEVICE_ID = 0;
 
@@ -30,12 +69,14 @@ public class Controller implements AsyncProcessor {
     private static final int POINTER_ID_MOUSE = -1;
 
     private static final ScheduledExecutorService EXECUTOR = Executors.newSingleThreadScheduledExecutor();
+    private ExecutorService startAppExecutor;
 
     private Thread thread;
 
     private UhidManager uhidManager;
 
-    private final Device device;
+    private final int displayId;
+    private final boolean supportsInputEvents;
     private final ControlChannel controlChannel;
     private final CleanUp cleanUp;
     private final DeviceMessageSender sender;
@@ -44,27 +85,102 @@ public class Controller implements AsyncProcessor {
 
     private final KeyCharacterMap charMap = KeyCharacterMap.load(KeyCharacterMap.VIRTUAL_KEYBOARD);
 
+    private final AtomicBoolean isSettingClipboard = new AtomicBoolean();
+
+    private final AtomicReference<DisplayData> displayData = new AtomicReference<>();
+    private final Object displayDataAvailable = new Object(); // condition variable
+
     private long lastTouchDown;
     private final PointersState pointersState = new PointersState();
     private final MotionEvent.PointerProperties[] pointerProperties = new MotionEvent.PointerProperties[PointersState.MAX_POINTERS];
     private final MotionEvent.PointerCoords[] pointerCoords = new MotionEvent.PointerCoords[PointersState.MAX_POINTERS];
 
-    private boolean keepPowerModeOff;
+    private boolean keepDisplayPowerOff;
 
-    public Controller(Device device, ControlChannel controlChannel, CleanUp cleanUp, boolean clipboardAutosync, boolean powerOn) {
-        this.device = device;
+    // Used for resetting video encoding on RESET_VIDEO message
+    private SurfaceCapture surfaceCapture;
+
+    public Controller(ControlChannel controlChannel, CleanUp cleanUp, Options options) {
+        this.displayId = options.getDisplayId();
         this.controlChannel = controlChannel;
         this.cleanUp = cleanUp;
-        this.clipboardAutosync = clipboardAutosync;
-        this.powerOn = powerOn;
+        this.clipboardAutosync = options.getClipboardAutosync();
+        this.powerOn = options.getPowerOn();
         initPointers();
         sender = new DeviceMessageSender(controlChannel);
+
+        supportsInputEvents = Device.supportsInputEvents(displayId);
+        if (!supportsInputEvents) {
+            Ln.w("Input events are not supported for secondary displays before Android 10");
+        }
+
+        if (clipboardAutosync) {
+            // If control and autosync are enabled, synchronize Android clipboard to the computer automatically
+            ClipboardManager clipboardManager = ServiceManager.getClipboardManager();
+            if (clipboardManager != null) {
+                clipboardManager.addPrimaryClipChangedListener(() -> {
+                    if (isSettingClipboard.get()) {
+                        // This is a notification for the change we are currently applying, ignore it
+                        return;
+                    }
+                    String text = Device.getClipboardText();
+                    if (text != null) {
+                        DeviceMessage msg = DeviceMessage.createClipboard(text);
+                        sender.send(msg);
+                    }
+                });
+            } else {
+                Ln.w("No clipboard manager, copy-paste between device and computer will not work");
+            }
+        }
+    }
+
+    @Override
+    public void onNewVirtualDisplay(int virtualDisplayId, PositionMapper positionMapper) {
+        DisplayData data = new DisplayData(virtualDisplayId, positionMapper);
+        DisplayData old = this.displayData.getAndSet(data);
+        if (old == null) {
+            // The very first time the Controller is notified of a new virtual display
+            synchronized (displayDataAvailable) {
+                displayDataAvailable.notify();
+            }
+        }
+    }
+
+    public void setSurfaceCapture(SurfaceCapture surfaceCapture) {
+        this.surfaceCapture = surfaceCapture;
     }
 
     private UhidManager getUhidManager() {
         if (uhidManager == null) {
-            uhidManager = new UhidManager(sender);
+            int uhidDisplayId = displayId;
+            if (Build.VERSION.SDK_INT >= AndroidVersions.API_35_ANDROID_15) {
+                if (displayId == Device.DISPLAY_ID_NONE) {
+                    // Mirroring a new virtual display id (using --new-display-id feature) on Android >= 15, where the UHID mouse pointer can be
+                    // associated to the virtual display
+                    try {
+                        // Wait for at most 1 second until a virtual display id is known
+                        DisplayData data = waitDisplayData(1000);
+                        if (data != null) {
+                            uhidDisplayId = data.virtualDisplayId;
+                        }
+                    } catch (InterruptedException e) {
+                        // do nothing
+                    }
+                }
+            }
+
+            String displayUniqueId = null;
+            if (uhidDisplayId > 0) {
+                // Ignore Device.DISPLAY_ID_NONE and 0 (main display)
+                DisplayInfo displayInfo = ServiceManager.getDisplayManager().getDisplayInfo(uhidDisplayId);
+                if (displayInfo != null) {
+                    displayUniqueId = displayInfo.getUniqueId();
+                }
+            }
+            uhidManager = new UhidManager(sender, displayUniqueId);
         }
+
         return uhidManager;
     }
 
@@ -84,8 +200,8 @@ public class Controller implements AsyncProcessor {
 
     private void control() throws IOException {
         // on start, power on the device
-        if (powerOn && !Device.isScreenOn()) {
-            device.pressReleaseKeycode(KeyEvent.KEYCODE_POWER, Device.INJECT_MODE_ASYNC);
+        if (powerOn && displayId == 0 && !Device.isScreenOn(displayId)) {
+            Device.pressReleaseKeycode(KeyEvent.KEYCODE_POWER, displayId, Device.INJECT_MODE_ASYNC);
 
             // dirty hack
             // After POWER is injected, the device is powered on asynchronously.
@@ -138,10 +254,6 @@ public class Controller implements AsyncProcessor {
         sender.join();
     }
 
-    public DeviceMessageSender getSender() {
-        return sender;
-    }
-
     private boolean handleEvent() throws IOException {
         ControlMessage msg;
         try {
@@ -153,27 +265,27 @@ public class Controller implements AsyncProcessor {
 
         switch (msg.getType()) {
             case ControlMessage.TYPE_INJECT_KEYCODE:
-                if (device.supportsInputEvents()) {
+                if (supportsInputEvents) {
                     injectKeycode(msg.getAction(), msg.getKeycode(), msg.getRepeat(), msg.getMetaState());
                 }
                 break;
             case ControlMessage.TYPE_INJECT_TEXT:
-                if (device.supportsInputEvents()) {
+                if (supportsInputEvents) {
                     injectText(msg.getText());
                 }
                 break;
             case ControlMessage.TYPE_INJECT_TOUCH_EVENT:
-                if (device.supportsInputEvents()) {
+                if (supportsInputEvents) {
                     injectTouch(msg.getAction(), msg.getPointerId(), msg.getPosition(), msg.getPressure(), msg.getActionButton(), msg.getButtons());
                 }
                 break;
             case ControlMessage.TYPE_INJECT_SCROLL_EVENT:
-                if (device.supportsInputEvents()) {
+                if (supportsInputEvents) {
                     injectScroll(msg.getPosition(), msg.getHScroll(), msg.getVScroll(), msg.getButtons());
                 }
                 break;
             case ControlMessage.TYPE_BACK_OR_SCREEN_ON:
-                if (device.supportsInputEvents()) {
+                if (supportsInputEvents) {
                     pressBackOrTurnScreenOn(msg.getAction());
                 }
                 break;
@@ -192,25 +304,16 @@ public class Controller implements AsyncProcessor {
             case ControlMessage.TYPE_SET_CLIPBOARD:
                 setClipboard(msg.getText(), msg.getPaste(), msg.getSequence());
                 break;
-            case ControlMessage.TYPE_SET_SCREEN_POWER_MODE:
-                if (device.supportsInputEvents()) {
-                    int mode = msg.getAction();
-                    boolean setPowerModeOk = Device.setScreenPowerMode(mode);
-                    if (setPowerModeOk) {
-                        keepPowerModeOff = mode == Device.POWER_MODE_OFF;
-                        Ln.i("Device screen turned " + (mode == Device.POWER_MODE_OFF ? "off" : "on"));
-                        if (cleanUp != null) {
-                            boolean mustRestoreOnExit = mode != Device.POWER_MODE_NORMAL;
-                            cleanUp.setRestoreNormalPowerMode(mustRestoreOnExit);
-                        }
-                    }
+            case ControlMessage.TYPE_SET_DISPLAY_POWER:
+                if (supportsInputEvents) {
+                    setDisplayPower(msg.getOn());
                 }
                 break;
             case ControlMessage.TYPE_ROTATE_DEVICE:
-                device.rotateDevice();
+                Device.rotateDevice(getActionDisplayId());
                 break;
             case ControlMessage.TYPE_UHID_CREATE:
-                getUhidManager().open(msg.getId(), msg.getText(), msg.getData());
+                getUhidManager().open(msg.getId(), msg.getVendorId(), msg.getProductId(), msg.getText(), msg.getData());
                 break;
             case ControlMessage.TYPE_UHID_INPUT:
                 getUhidManager().writeInput(msg.getId(), msg.getData());
@@ -221,6 +324,12 @@ public class Controller implements AsyncProcessor {
             case ControlMessage.TYPE_OPEN_HARD_KEYBOARD_SETTINGS:
                 openHardKeyboardSettings();
                 break;
+            case ControlMessage.TYPE_START_APP:
+                startAppAsync(msg.getText());
+                break;
+            case ControlMessage.TYPE_RESET_VIDEO:
+                resetVideo();
+                break;
             default:
                 // do nothing
         }
@@ -229,10 +338,11 @@ public class Controller implements AsyncProcessor {
     }
 
     private boolean injectKeycode(int action, int keycode, int repeat, int metaState) {
-        if (keepPowerModeOff && action == KeyEvent.ACTION_UP && (keycode == KeyEvent.KEYCODE_POWER || keycode == KeyEvent.KEYCODE_WAKEUP)) {
-            schedulePowerModeOff();
+        if (keepDisplayPowerOff && action == KeyEvent.ACTION_UP && (keycode == KeyEvent.KEYCODE_POWER || keycode == KeyEvent.KEYCODE_WAKEUP)) {
+            assert displayId != Device.DISPLAY_ID_NONE;
+            scheduleDisplayPowerOff(displayId);
         }
-        return device.injectKeyEvent(action, keycode, repeat, metaState, Device.INJECT_MODE_ASYNC);
+        return injectKeyEvent(action, keycode, repeat, metaState, Device.INJECT_MODE_ASYNC);
     }
 
     private boolean injectChar(char c) {
@@ -242,8 +352,10 @@ public class Controller implements AsyncProcessor {
         if (events == null) {
             return false;
         }
+
+        int actionDisplayId = getActionDisplayId();
         for (KeyEvent event : events) {
-            if (!device.injectEvent(event, Device.INJECT_MODE_ASYNC)) {
+            if (!Device.injectEvent(event, actionDisplayId, Device.INJECT_MODE_ASYNC)) {
                 return false;
             }
         }
@@ -262,15 +374,47 @@ public class Controller implements AsyncProcessor {
         return successCount;
     }
 
+    private Pair<Point, Integer> getEventPointAndDisplayId(Position position) {
+        // it hides the field on purpose, to read it with atomic access
+        @SuppressWarnings("checkstyle:HiddenField")
+        DisplayData displayData = this.displayData.get();
+        // In scrcpy, displayData should never be null (a touch event can only be generated from the client when a video frame is present).
+        // However, it is possible to send events without video playback when using scrcpy-server alone (except for virtual displays).
+        assert displayData != null || displayId != Device.DISPLAY_ID_NONE : "Cannot receive a positional event without a display";
+
+        Point point;
+        int targetDisplayId;
+        if (displayData != null) {
+            point = displayData.positionMapper.map(position);
+            if (point == null) {
+                if (Ln.isEnabled(Ln.Level.VERBOSE)) {
+                    Size eventSize = position.getScreenSize();
+                    Size currentSize = displayData.positionMapper.getVideoSize();
+                    Ln.v("Ignore positional event generated for size " + eventSize + " (current size is " + currentSize + ")");
+                }
+                return null;
+            }
+            targetDisplayId = displayData.virtualDisplayId;
+        } else {
+            // No display, use the raw coordinates
+            point = position.getPoint();
+            targetDisplayId = displayId;
+        }
+
+        return Pair.create(point, targetDisplayId);
+    }
+
     private boolean injectTouch(int action, long pointerId, Position position, float pressure, int actionButton, int buttons) {
         long now = SystemClock.uptimeMillis();
 
-        Point point = device.getPhysicalPoint(position);
-        if (point == null) {
-            Ln.w("Ignore touch event, it was generated for a different device size");
+        Pair<Point, Integer> pair = getEventPointAndDisplayId(position);
+        if (pair == null) {
             return false;
         }
 
+        Point point = pair.first;
+        int targetDisplayId = pair.second;
+
         int pointerIndex = pointersState.getPointerIndex(pointerId);
         if (pointerIndex == -1) {
             Ln.w("Too many pointers for touch event");
@@ -318,13 +462,13 @@ public class Controller implements AsyncProcessor {
          *
          * Otherwise, Chrome does not work properly: <https://github.com/Genymobile/scrcpy/issues/3635>
          */
-        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && source == InputDevice.SOURCE_MOUSE) {
+        if (Build.VERSION.SDK_INT >= AndroidVersions.API_23_ANDROID_6_0 && source == InputDevice.SOURCE_MOUSE) {
             if (action == MotionEvent.ACTION_DOWN) {
                 if (actionButton == buttons) {
                     // First button pressed: ACTION_DOWN
                     MotionEvent downEvent = MotionEvent.obtain(lastTouchDown, now, MotionEvent.ACTION_DOWN, pointerCount, pointerProperties,
                             pointerCoords, 0, buttons, 1f, 1f, DEFAULT_DEVICE_ID, 0, source, 0);
-                    if (!device.injectEvent(downEvent, Device.INJECT_MODE_ASYNC)) {
+                    if (!Device.injectEvent(downEvent, targetDisplayId, Device.INJECT_MODE_ASYNC)) {
                         return false;
                     }
                 }
@@ -335,7 +479,7 @@ public class Controller implements AsyncProcessor {
                 if (!InputManager.setActionButton(pressEvent, actionButton)) {
                     return false;
                 }
-                if (!device.injectEvent(pressEvent, Device.INJECT_MODE_ASYNC)) {
+                if (!Device.injectEvent(pressEvent, targetDisplayId, Device.INJECT_MODE_ASYNC)) {
                     return false;
                 }
 
@@ -349,7 +493,7 @@ public class Controller implements AsyncProcessor {
                 if (!InputManager.setActionButton(releaseEvent, actionButton)) {
                     return false;
                 }
-                if (!device.injectEvent(releaseEvent, Device.INJECT_MODE_ASYNC)) {
+                if (!Device.injectEvent(releaseEvent, targetDisplayId, Device.INJECT_MODE_ASYNC)) {
                     return false;
                 }
 
@@ -357,7 +501,7 @@ public class Controller implements AsyncProcessor {
                     // Last button released: ACTION_UP
                     MotionEvent upEvent = MotionEvent.obtain(lastTouchDown, now, MotionEvent.ACTION_UP, pointerCount, pointerProperties,
                             pointerCoords, 0, buttons, 1f, 1f, DEFAULT_DEVICE_ID, 0, source, 0);
-                    if (!device.injectEvent(upEvent, Device.INJECT_MODE_ASYNC)) {
+                    if (!Device.injectEvent(upEvent, targetDisplayId, Device.INJECT_MODE_ASYNC)) {
                         return false;
                     }
                 }
@@ -368,17 +512,20 @@ public class Controller implements AsyncProcessor {
 
         MotionEvent event = MotionEvent.obtain(lastTouchDown, now, action, pointerCount, pointerProperties, pointerCoords, 0, buttons, 1f, 1f,
                 DEFAULT_DEVICE_ID, 0, source, 0);
-        return device.injectEvent(event, Device.INJECT_MODE_ASYNC);
+        return Device.injectEvent(event, targetDisplayId, Device.INJECT_MODE_ASYNC);
     }
 
     private boolean injectScroll(Position position, float hScroll, float vScroll, int buttons) {
         long now = SystemClock.uptimeMillis();
-        Point point = device.getPhysicalPoint(position);
-        if (point == null) {
-            // ignore event
+
+        Pair<Point, Integer> pair = getEventPointAndDisplayId(position);
+        if (pair == null) {
             return false;
         }
 
+        Point point = pair.first;
+        int targetDisplayId = pair.second;
+
         MotionEvent.PointerProperties props = pointerProperties[0];
         props.id = 0;
 
@@ -390,22 +537,22 @@ public class Controller implements AsyncProcessor {
 
         MotionEvent event = MotionEvent.obtain(lastTouchDown, now, MotionEvent.ACTION_SCROLL, 1, pointerProperties, pointerCoords, 0, buttons, 1f, 1f,
                 DEFAULT_DEVICE_ID, 0, InputDevice.SOURCE_MOUSE, 0);
-        return device.injectEvent(event, Device.INJECT_MODE_ASYNC);
+        return Device.injectEvent(event, targetDisplayId, Device.INJECT_MODE_ASYNC);
     }
 
     /**
-     * Schedule a call to set power mode to off after a small delay.
+     * Schedule a call to set display power to off after a small delay.
      */
-    private static void schedulePowerModeOff() {
+    private static void scheduleDisplayPowerOff(int displayId) {
         EXECUTOR.schedule(() -> {
-            Ln.i("Forcing screen off");
-            Device.setScreenPowerMode(Device.POWER_MODE_OFF);
+            Ln.i("Forcing display off");
+            Device.setDisplayPower(displayId, false);
         }, 200, TimeUnit.MILLISECONDS);
     }
 
     private boolean pressBackOrTurnScreenOn(int action) {
-        if (Device.isScreenOn()) {
-            return device.injectKeyEvent(action, KeyEvent.KEYCODE_BACK, 0, 0, Device.INJECT_MODE_ASYNC);
+        if (displayId == Device.DISPLAY_ID_NONE || Device.isScreenOn(displayId)) {
+            return injectKeyEvent(action, KeyEvent.KEYCODE_BACK, 0, 0, Device.INJECT_MODE_ASYNC);
         }
 
         // Screen is off
@@ -415,18 +562,19 @@ public class Controller implements AsyncProcessor {
             return true;
         }
 
-        if (keepPowerModeOff) {
-            schedulePowerModeOff();
+        if (keepDisplayPowerOff) {
+            assert displayId != Device.DISPLAY_ID_NONE;
+            scheduleDisplayPowerOff(displayId);
         }
-        return device.pressReleaseKeycode(KeyEvent.KEYCODE_POWER, Device.INJECT_MODE_ASYNC);
+        return pressReleaseKeycode(KeyEvent.KEYCODE_POWER, Device.INJECT_MODE_ASYNC);
     }
 
     private void getClipboard(int copyKey) {
         // On Android >= 7, press the COPY or CUT key if requested
-        if (copyKey != ControlMessage.COPY_KEY_NONE && Build.VERSION.SDK_INT >= Build.VERSION_CODES.N && device.supportsInputEvents()) {
+        if (copyKey != ControlMessage.COPY_KEY_NONE && Build.VERSION.SDK_INT >= AndroidVersions.API_24_ANDROID_7_0 && supportsInputEvents) {
             int key = copyKey == ControlMessage.COPY_KEY_COPY ? KeyEvent.KEYCODE_COPY : KeyEvent.KEYCODE_CUT;
             // Wait until the event is finished, to ensure that the clipboard text we read just after is the correct one
-            device.pressReleaseKeycode(key, Device.INJECT_MODE_WAIT_FOR_FINISH);
+            pressReleaseKeycode(key, Device.INJECT_MODE_WAIT_FOR_FINISH);
         }
 
         // If clipboard autosync is enabled, then the device clipboard is synchronized to the computer clipboard whenever it changes, in
@@ -442,14 +590,16 @@ public class Controller implements AsyncProcessor {
     }
 
     private boolean setClipboard(String text, boolean paste, long sequence) {
-        boolean ok = device.setClipboardText(text);
+        isSettingClipboard.set(true);
+        boolean ok = Device.setClipboardText(text);
+        isSettingClipboard.set(false);
         if (ok) {
             Ln.i("Device clipboard set");
         }
 
         // On Android >= 7, also press the PASTE key if requested
-        if (paste && Build.VERSION.SDK_INT >= Build.VERSION_CODES.N && device.supportsInputEvents()) {
-            device.pressReleaseKeycode(KeyEvent.KEYCODE_PASTE, Device.INJECT_MODE_ASYNC);
+        if (paste && Build.VERSION.SDK_INT >= AndroidVersions.API_24_ANDROID_7_0 && supportsInputEvents) {
+            pressReleaseKeycode(KeyEvent.KEYCODE_PASTE, Device.INJECT_MODE_ASYNC);
         }
 
         if (sequence != ControlMessage.SEQUENCE_INVALID) {
@@ -465,4 +615,142 @@ public class Controller implements AsyncProcessor {
         Intent intent = new Intent("android.settings.HARD_KEYBOARD_SETTINGS");
         ServiceManager.getActivityManager().startActivity(intent);
     }
+
+    private boolean injectKeyEvent(int action, int keyCode, int repeat, int metaState, int injectMode) {
+        return Device.injectKeyEvent(action, keyCode, repeat, metaState, getActionDisplayId(), injectMode);
+    }
+
+    private boolean pressReleaseKeycode(int keyCode, int injectMode) {
+        return Device.pressReleaseKeycode(keyCode, getActionDisplayId(), injectMode);
+    }
+
+    private int getActionDisplayId() {
+        if (displayId != Device.DISPLAY_ID_NONE) {
+            // Real screen mirrored, use the source display id
+            return displayId;
+        }
+
+        // Virtual display created by --new-display, use the virtualDisplayId
+        DisplayData data = displayData.get();
+        if (data == null) {
+            // If no virtual display id is initialized yet, use the main display id
+            return 0;
+        }
+
+        return data.virtualDisplayId;
+    }
+
+    private void startAppAsync(String name) {
+        if (startAppExecutor == null) {
+            startAppExecutor = Executors.newSingleThreadExecutor();
+        }
+
+        // Listing and selecting the app may take a lot of time
+        startAppExecutor.submit(() -> startApp(name));
+    }
+
+    private void startApp(String name) {
+        boolean forceStopBeforeStart = name.startsWith("+");
+        if (forceStopBeforeStart) {
+            name = name.substring(1);
+        }
+
+        DeviceApp app;
+        boolean searchByName = name.startsWith("?");
+        if (searchByName) {
+            name = name.substring(1);
+
+            Ln.i("Processing Android apps... (this may take some time)");
+            List<DeviceApp> apps = Device.findByName(name);
+            if (apps.isEmpty()) {
+                Ln.w("No app found for name \"" + name + "\"");
+                return;
+            }
+
+            if (apps.size() > 1) {
+                String title = "No unique app found for name \"" + name + "\":";
+                Ln.w(LogUtils.buildAppListMessage(title, apps));
+                return;
+            }
+
+            app = apps.get(0);
+        } else {
+            app = Device.findByPackageName(name);
+            if (app == null) {
+                Ln.w("No app found for package \"" + name + "\"");
+                return;
+            }
+        }
+
+        int startAppDisplayId = getStartAppDisplayId();
+        if (startAppDisplayId == Device.DISPLAY_ID_NONE) {
+            Ln.e("No known display id to start app \"" + name + "\"");
+            return;
+        }
+
+        Ln.i("Starting app \"" + app.getName() + "\" [" + app.getPackageName() + "] on display " + startAppDisplayId + "...");
+        Device.startApp(app.getPackageName(), startAppDisplayId, forceStopBeforeStart);
+    }
+
+    private int getStartAppDisplayId() {
+        if (displayId != Device.DISPLAY_ID_NONE) {
+            return displayId;
+        }
+
+        // Mirroring a new virtual display id (using --new-display-id feature)
+        try {
+            // Wait for at most 1 second until a virtual display id is known
+            DisplayData data = waitDisplayData(1000);
+            if (data != null) {
+                return data.virtualDisplayId;
+            }
+        } catch (InterruptedException e) {
+            // do nothing
+        }
+
+        // No display id available
+        return Device.DISPLAY_ID_NONE;
+    }
+
+    private DisplayData waitDisplayData(long timeoutMillis) throws InterruptedException {
+        long deadline = System.currentTimeMillis() + timeoutMillis;
+
+        synchronized (displayDataAvailable) {
+            DisplayData data = displayData.get();
+            while (data == null) {
+                long timeout = deadline - System.currentTimeMillis();
+                if (timeout < 0) {
+                    return null;
+                }
+                if (timeout > 0) {
+                    displayDataAvailable.wait(timeout);
+                }
+                data = displayData.get();
+            }
+
+            return data;
+        }
+    }
+
+    private void setDisplayPower(boolean on) {
+        // Change the power of the main display when mirroring a virtual display
+        int targetDisplayId = displayId != Device.DISPLAY_ID_NONE ? displayId : 0;
+        boolean setDisplayPowerOk = Device.setDisplayPower(targetDisplayId, on);
+        if (setDisplayPowerOk) {
+            // Do not keep display power off for virtual displays: MOD+p must wake up the physical device
+            keepDisplayPowerOff = displayId != Device.DISPLAY_ID_NONE && !on;
+            Ln.i("Device display turned " + (on ? "on" : "off"));
+            if (cleanUp != null) {
+                boolean mustRestoreOnExit = !on;
+                cleanUp.setRestoreDisplayPower(mustRestoreOnExit);
+            }
+        }
+    }
+
+    private void resetVideo() {
+        if (surfaceCapture != null) {
+            Ln.i("Video capture reset");
+            surfaceCapture.requestInvalidate();
+        }
+    }
 }
